function [Reconstructed, Fidelity, data] = simObrien_v2( U, sgnA22, singles, HOMs, maxHOM)
% insert a square unitary matrix U and get a reconstructed matrix back
% this method is based on A. Laing and J. O'brien ArXiv paper
% on superstable tomography of any linear optical network
%
% This function is adapted from the simObrien_imperfections simulation file
% but in here, all noise terms for simulations have been omitted
%
% Additionally, this file can be used to process experimental data in the
% input is (partially) provided. If the data is partially provided, please
% fill in the correct matrix U so that the program can generate the missing
% information (single counts or HOM visibilities)
%
% The program corrects for imperfect photons (partial indistinguishability)
% by correcting for the maximal possible HOM dip.
%
% If the reconstructed matrix is not unitary, a polar decomposition will be
% performed to find the closest possible unitary matrix (closest by what
% metric?). If it fails miserably (NaN elements), the given matrix U will
% be returned, together with a warning.
%
% Input:
%   U           the square unitary matrix
%   
%  optional
%   sgnA22      the sign of the phase of U(2,2) is ambiguous. Any positive
%               (negative) number will result in the program picking a
%               positive (negative) phase sign. Default: positive
%   singles     A nxn matrix of real numbers corresponding to the single
%               counts. It does not need to be corrected for known
%               coupling-losses. The numbers do not need to be normalized
%   HOMs        a 2xnxn matrix of the HOM visibilities. More information
%               follows later (to do)
%   maxHOM      maximum HOM visibility (1 for perfect photons)
%
%
% Output:
%   Reconstructed   The reconstructed matrix U (unless reconstruction fails
%
% optional:
%   Fidelity        The fidelity of the matrix with the target matrix U
%   data            the simulation / experimental data (singles/HOM
%                   visibilities etc.)
% 
%  To do: proper input handeling

% rephase the matrix (no phase @ first col/row)
U = rephase(U);


% input validation
n = size(U,1);
if size(U,1) ~= size(U,2)
    error('Matrix should be square')
end


% simulate (noisy) measurement data
if nargin == 1
    % alpha(2,2) is always positive (assumption of protocol)
    sgnA22 = 1;
    data = getSimData( U);
    
elseif nargin == 2
    % now user defined what sign (s)he wants for alpha_2_2
    data = getSimData( U);
    
elseif nargin ==3
    % use the user defined sign and single counts
    % first generate the data based on the ideal matrix
    data = getSimData(U);
    % now overwrite with the actual measured data
    data.R = singles;

elseif nargin == 4
    % now also with (partial) experimetal data
    % first generate the data based on the ideal matrix
    data = getSimData(U);
    % now overwrite with the actual measured data
    data.Visibility = HOMs;
    data.R = singles;
    
elseif nargin == 5
    data = getSimData( U);
    data.Visibility = HOMs;
    data.R = singles;
    data.MaxHOM = maxHOM;
else
   error('Incorrect # input parameters given')
end 



% allocate some memory
res = initRes( n);
res.usedHOM = zeros(n,n,2,2);

%% start of the actual protocol
% go through the innerpart of the matrix
for gg = 2 : n
    for hh = 2 : n
        
        % eq. 5
        jj = 1;
        kk = 1;
        
        % get the x,y
        res = getX_Y( data, res, gg, hh, jj, kk);
        
        % get the phase alpha
        res = getAlpha( data, res, gg, hh, jj, kk);
        
        
        % element (2,2)
        if gg == 2 && hh == 2
            % choice: first phase positive or negative
            if sgnA22 > 0
                res.alpha(gg,hh) = abs( res.alpha(gg,hh));
            else
                res.alpha(gg,hh) =-abs( res.alpha(gg,hh));
            end
            
        else
            if gg == 2 && hh ~= 2
                % row 2
                jj = 1;
                kk = 2;
                
            elseif gg ~= 2 && hh == 2
                % column 2
                jj = 2;
                kk = 1;
                
            elseif gg ~= 2 && hh ~= 2
                % somewhere inside (not 2nd row || not 2nd column)
                jj = 2;
                kk = 2;
            end
            
            % get new y
            res = getX_Y( data, res, gg, hh, jj, kk);
            
            % fix alpha sign
            [res, sgn] = getAlphaSign( res, data, gg, hh, jj, kk);
            res.alpha(gg,hh) = res.alpha(gg,hh) * sgn;
        end
    end
end

% now solve the linear set of equations (eq. 9)
res = solveMatrix_tau( res);


% get the remaining tau
res = fillTau( res);


% finalize the reconstruction
Reconstructed = res.tau .* exp( 1i *res.alpha);

% append missing info to data (in case the user wants to inspect later)
data.Reconstructed = Reconstructed;
data.usedHOM = res.usedHOM;

% check if the matrix is unitary
% if not, do a polar decomposition
Reconstructed = doPolarD( Reconstructed);


% compute the fidelity
[Fidelity, Reconstructed] = getFidelity( U, Reconstructed);


% check against NaN and Inf problems
nNan = sum( sum( isnan(Reconstructed)));
nInf = sum( sum( isinf(Reconstructed)));
if nNan >0
    warning('Reconstructed matrix contains NaN, return ideal matrix U')
    Reconstructed = U;
elseif nInf > 0
    warning('Reconstructed matrix contains infinity, return ideal matrix U')
    Reconstructed = U;
end



%% FUNCTION SPACE
function U = rephase(U)
%U = REPHASE(U)
%   Sets the phases of the first column and row to zero by left and right
%   multiplication of diagonal matrices.

% Input validation
s = size(U);
if length(s) ~= 2
    error('U must be 2 dimensional matrix')
end
if s(1) ~= s(2)
    error('U must be a square matrix')
end

% Rephase row 1
D1 = diag(exp(-1i*angle(U(1,:))));
U = U*D1;

% Rephase column 1
D2 = diag(exp(-1i*angle(U(:,1))));
U = D2*U;

% guarantee a positive phase @ element (2,2)
% don't do this as this is problematic. It will make the U(2,2) angle
% positive, however the top row is not angle = 0 anymore
% if angle(U(2,2))<0
%     U = abs(U) * exp( -1i* angle(U));
% end
end

function data = getSimData( U)
% simulate the input parameters (rates & visibility)
% include actual input/output losses, darkcounts etc.
% store all data in 2 matrices: Rates, Visibility
% Visibility is a multidimensional matrix

% disp('simulate the data')
data.nModes = length(U);
data.maxHOM = 1;


% get the counts
data.R = abs(U).^2;

% the method fails when 0 rates are observed
if min(min(data.R)) == 0
%     disp('There are matrix elements with amplitude 0!')
    data.R = data.R*1e6+1;
end

% now get the HOM visibilities
data.Visibility = zeros( data.nModes, data.nModes, 2, 2);
problem = 0;
for ggg = 2 : data.nModes
    
    for hhh = 2 : data.nModes
        
        for jjj = 1 : 2
            
            for kkk = 1 : 2
                Usub = U([jjj,ggg],[kkk,hhh]);
                % ideal case
                C = permanentRyser( abs( Usub) .^2);
                Q = abs( permanentRyser( Usub)).^2;
                % in case there are no counts
                if C > 0
                    data.Visibility(ggg,hhh,jjj,kkk) = (C-Q)/C;
%                     problem = problem+1;
                end
            end
        end
    end
end
% if |u| == 0 --> cannot determine the HOM visibility
if problem == 1
%     disp(['# HOM dips which couldnt be determined (=NaN): ',num2str(problem)])
end
end

function res = initRes( nModes)
% allocate memory
res.nModes  = nModes;
res.alpha   = zeros( nModes);
res.x       = zeros( nModes, nModes, 2, 2);
res.y       = zeros( nModes, nModes, 2, 2);
res.Matrix  =  ones( nModes);
res.tau     = zeros( nModes);
end

function res = getX_Y( data, res, g, h, j, k)
% get x and y (eq. 5 and eq. 6) from the Rates (data)
% for the given indices g h j k
X = data.R(j,k) *data.R(g,h) /data.R(j,h) /data.R(g,k);

res.x(g, h, j, k) = sqrt(X);
res.y(g, h, j, k) = sqrt(X) + 1/sqrt(X);


end

function res = getAlpha( data, res, g, h, j, k)
% get |alpha| from eq. 6
% optional: divide by known maxHOM for the reconstruction
% arg = -0.5 *data.Visibility(g,h,j,k) *res.y(g,h,j,k);
arg = -0.5 *data.Visibility(g,h,j,k) *res.y(g,h,j,k)/data.maxHOM;

res.usedHOM(g,h,j,k) = 1;

% hard code the bounds of the acos function, prevent it to go imaginary
if arg > 1
    arg = 1;
elseif arg < -1
    arg = -1;
end

res.alpha(g,h) = wrapToPi(acos( arg)...
    -res.alpha(j,k) +res.alpha(j,h) +res.alpha(g,k));

% sometimes noise(?) makes it imaginary, so deal with that then
if imag(res.alpha(g,h)) ~= 0
    warning('imaginary phase!!')
    if imag(res.alpha(g,h)) <0
        res.alpha(g,h) = wrapToPi(-1-res.alpha(j,k) +res.alpha(j,h) +res.alpha(g,k));
    else
        res.alpha(g,h) =  wrapToPi(1-res.alpha(j,k) +res.alpha(j,h) +res.alpha(g,k));
    end
    
    res.alpha(g,h) = imag(res.alpha(g,h));
end
end

function [res, sgn] = getAlphaSign( res, data, g, h, j, k)
% get the sign of alpha (eq. 7)
beta = abs( acos( -0.5 *data.Visibility(g,h,j,k) /data.maxHOM...
                  *res.y(g,h,j,k)));

res.usedHOM(g,h,j,k) = 1;

term1 = abs( beta -abs( wrapToPi(...
    res.alpha(j,k) -res.alpha(j,h) -res.alpha(g,k) -abs(res.alpha(g,h)))));
term2 = abs( beta -abs( wrapToPi(...
    res.alpha(j,k) -res.alpha(j,h) -res.alpha(g,k) +abs(res.alpha(g,h)))));
sgn = sign( term1 - term2);

end

function res = solveMatrix_tau( res)
% eq 9
% you solve with the normal x_g,h,j,k, the resulting row/column are placed
% in the matrix with the x_tilde_g,h,j,k

% fill up the matrix M_mu
ind = 2 : res.nModes;
res.Matrix(ind,ind) = res.x(ind,ind,1,1) .*exp( 1i * res.alpha(ind,ind));

% use a linear solver to solve the equations
tau_col = abs( sqrt( linsolve( res.Matrix',[1, zeros(1,res.nModes-1)]')));
tau_row = abs( sqrt( linsolve( res.Matrix, [1, zeros(1,res.nModes-1)]')));

res.tau(1,:) = tau_row;
res.tau(:,1) = tau_col;
end

function res = fillTau( res)
% fills up the still unknown values of tau in the inside part of the matrix
% after solving for the first row/col
% uses eq. 5 ( x = tau *tau / tau / tau)

j = 1;
k = 1;

% loop through the inside
for g = 2 : res.nModes
    for h = 2 : res.nModes
        % rewrite lower line of eq. 5
        res.tau(g,h) = res.x(g,h,j,k)...
            *res.tau(j,h) *res.tau(g,k) /res.tau(j,k);
    end
end
end

function R = doPolarD( R)
    margin = 1E-3;
    nn = length(R);
    if abs(max(max(R*R' - eye(nn)))) > margin
        disp('Non unitary matrix found: do a polar decomposition')
        [A,~,B] = svd(R);
        R = A*B';
        
        % reset the first row/col to phase 0
        R = rephase(R);
    end
end

function [Fidelity,B] = getFidelity( A, B)
% computes the fidelity as defined in the Burgwal Imperfections paper
% (Optics Express 2017)
% note that the phase sign can be ambigious, so if the fidelity is too
% small, simply try to flip the phases

m = length(A);
F1 = 1 / m * trace( abs(A'*B));

% in a bad outcome: try flipping the phases
F2 = 0;
R2 = abs(B) .*exp(-1i*angle(B));
F2 = 1 / m * trace( abs(A'*B));

% take the best value
if F1 > F2
    Fidelity = F1;
else
    Fidelity = F2;
    B = R2;
end

end

end